"""fix_parse_websearch.

Revision ID: 19cdd3ab86f6
Revises: 6896a54e9483
Create Date: 2021-11-17 21:23:09.959694

"""
from alembic import op
from sqlalchemy import text

# revision identifiers, used by Alembic.
revision = "19cdd3ab86f6"
down_revision = "6896a54e9483"
branch_labels = None
depends_on = None


def upgrade() -> None:
    command = """
              CREATE OR REPLACE FUNCTION parse_websearch(config regconfig, search_query text)
              RETURNS tsquery AS $$
              SELECT
                  string_agg(
                      (
                          CASE
                              WHEN position('''' IN words.word) > 0 THEN CONCAT(words.word, ':*')
                              ELSE words.word
                          END
                      ),
                      ' '
                  )::tsquery
              FROM (
                  SELECT trim(
                      regexp_split_to_table(
                          websearch_to_tsquery(config, lower(search_query))::text,
                          ' '
                      )
                  ) AS word
              ) AS words
              $$ LANGUAGE SQL IMMUTABLE;

              CREATE OR REPLACE FUNCTION parse_websearch(search_query text)
              RETURNS tsquery AS $$
              SELECT parse_websearch('pg_catalog.simple', search_query);
              $$ LANGUAGE SQL IMMUTABLE;"""
    op.execute(text(command))


def downgrade() -> None:
    op.execute(text("DROP FUNCTION public.parse_websearch(regconfig, text);"))
    op.execute(text("DROP FUNCTION public.parse_websearch(text);"))



from itertools import count
from typing import Any, Generator, Optional
from uuid import UUID

import strawberry
from pydantic.utils import to_lower_camel
from strawberry.scalars import JSON

from orchestrator.domain.base import SubscriptionModel
from orchestrator.services.subscriptions import build_extended_domain_model


@strawberry.type
class ProductBlockInstance:
    id: int
    parent: Optional[int]
    subscription_instance_id: UUID
    owner_subscription_id: UUID
    product_block_instance_values: JSON

    @strawberry.field(description="Returns all resource types of a product block", deprecation_reason="changed to product_block_instance_values")  # type: ignore
    async def resource_types(self) -> JSON:
        return {v["field"]: v["value"] for v in self.product_block_instance_values}


def is_product_block(candidate: Any) -> bool:
    if isinstance(candidate, dict):
        # TODO: also filter on tag (needs addition of tag in orchestrator endpoint)
        # NOTE: this crosses subscription boundaries. If needed we can add an additional filter to limit that.
        return candidate.get("owner_subscription_id", None)
    return False


def get_all_product_blocks(subscription: dict[str, Any], _tags: Optional[list[str]]) -> list[dict[str, Any]]:
    gen_id = count()

    def locate_product_block(candidate: dict[str, Any]) -> Generator:
        def new_product_block(item: dict[str, Any]) -> Generator:
            enriched_item = item | {"id": next(gen_id), "parent": candidate.get("id")}
            yield enriched_item
            yield from locate_product_block(enriched_item)

        for value in candidate.values():
            if is_product_block(value):
                yield from new_product_block(value)
            elif isinstance(value, list):
                for item in value:
                    if is_product_block(item):
                        yield from new_product_block(item)

    return list(locate_product_block(subscription))


async def get_subscription_product_blocks(
    subscription_id: UUID, tags: Optional[list[str]] = None, product_block_instance_values: Optional[list[str]] = None
) -> list[ProductBlockInstance]:
    subscription_model = SubscriptionModel.from_subscription(subscription_id)
    subscription = build_extended_domain_model(subscription_model)

    def to_product_block(product_block: dict[str, Any]) -> ProductBlockInstance:
        def is_resource_type(candidate: Any) -> bool:
            return isinstance(candidate, (bool, str, int, float, type(None)))

        def requested_resource_type(key: str) -> bool:
            return not product_block_instance_values or key in product_block_instance_values

        def included(key: str, value: Any) -> bool:
            return is_resource_type(value) and requested_resource_type(key) and key not in ("id", "parent")

        return ProductBlockInstance(
            id=product_block["id"],
            parent=product_block.get("parent"),
            owner_subscription_id=product_block["owner_subscription_id"],
            subscription_instance_id=product_block["subscription_instance_id"],
            product_block_instance_values=[
                {"field": to_lower_camel(k), "value": v} for k, v in product_block.items() if included(k, v)
            ],
        )

    product_blocks = (to_product_block(product_block) for product_block in get_all_product_blocks(subscription, tags))
    return [product_block for product_block in product_blocks if product_block.product_block_instance_values]



# Copyright 2019-2020 SURF.
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from http import HTTPStatus
from typing import List, Optional
from uuid import UUID

from fastapi.param_functions import Body
from fastapi.routing import APIRouter
from sqlalchemy.orm import joinedload, selectinload

from orchestrator.api.error_handling import raise_status
from orchestrator.api.models import delete, save, update
from orchestrator.db import ProductBlockTable, ProductTable, SubscriptionTable
from orchestrator.domain.lifecycle import ProductLifecycle
from orchestrator.schemas import ProductCRUDSchema, ProductSchema
from orchestrator.services.products import get_tags, get_types

router = APIRouter()


@router.get("/", response_model=List[ProductSchema])
def fetch(tag: Optional[str] = None, product_type: Optional[str] = None) -> List[ProductSchema]:
    query = ProductTable.query.options(
        selectinload(ProductTable.workflows),
        selectinload(ProductTable.fixed_inputs),
        selectinload(ProductTable.product_blocks).selectinload(ProductBlockTable.resource_types),
    )
    if tag:
        query = query.filter(ProductTable.__dict__["tag"] == tag)
    if product_type:
        query = query.filter(ProductTable.__dict__["product_type"] == product_type)

    return query.all()


@router.get("/{product_id}", response_model=ProductSchema)
def product_by_id(product_id: UUID) -> ProductTable:
    product = (
        ProductTable.query.options(
            joinedload(ProductTable.fixed_inputs),
            joinedload(ProductTable.product_blocks),
            joinedload(ProductTable.workflows),
        )
        .filter_by(product_id=product_id)
        .first()
    )
    if not product:
        raise_status(HTTPStatus.NOT_FOUND, f"Product id {product_id} not found")
    return product


@router.post("/", response_model=None, status_code=HTTPStatus.NO_CONTENT)
def save_product(data: ProductCRUDSchema = Body(...)) -> None:
    return save(ProductTable, data)


@router.put("/", response_model=None, status_code=HTTPStatus.NO_CONTENT)
def update_product(data: ProductCRUDSchema = Body(...)) -> None:
    return update(ProductTable, data)


@router.delete("/{product_id}", response_model=None, status_code=HTTPStatus.NO_CONTENT)
def delete_product(product_id: UUID) -> None:
    subscriptions = SubscriptionTable.query.filter(SubscriptionTable.product_id == product_id).all()
    if len(subscriptions) > 0:
        error_subscriptions = ", ".join(map(lambda sub: sub.description, subscriptions))
        raise_status(HTTPStatus.BAD_REQUEST, f"Product {product_id} is used in Subscriptions: {error_subscriptions}")
    return delete(ProductTable, product_id)


@router.get("/tags/all", response_model=List[str])
def tags() -> List[str]:
    return get_tags()


@router.get("/types/all", response_model=List[str])
def types() -> List[str]:
    return get_types()


@router.get("/statuses/all", response_model=List[ProductLifecycle])
def statuses() -> List[ProductLifecycle]:
    return ProductLifecycle.values()



"""customer_id to VARCHAR.

Revision ID: 165303a20fb1
Revises: a09ac125ea73
Create Date: 2023-07-17 13:53:23.932681

"""
from pathlib import Path

import sqlalchemy as sa
from alembic import op

revision = "165303a20fb1"
down_revision = "a09ac125ea73"
branch_labels = None
depends_on = None


def upgrade() -> None:
    conn = op.get_bind()

    revision_file_path = Path(__file__)
    with open(revision_file_path.with_suffix(".sql")) as f:
        conn.execute(sa.text(f.read()))


def downgrade() -> None:
    """This migration is irreversible!

    Once the type of `subscriptions.customer_id` has been changed
    from UUID to VARCHAR, it is not a failsafe operation to convert whatever value `customer_id` might now hold
    into a valid UUID type.

    In future, it will be necessary for downstream users to implement their own schema & data migrations
    if they want to (or even feasibly can) change the type of the `customer_id` column.
    """
    pass



from sqlalchemy.inspection import inspect

from orchestrator.db import ResourceTypeTable
from orchestrator.db.sorting.sorting import generic_column_sort, generic_sort
from orchestrator.utils.helpers import to_camel

RESOURCE_TYPE_SORT_FUNCTIONS_BY_COLUMN = {
    to_camel(key): generic_column_sort(value) for key, value in inspect(ResourceTypeTable).columns.items()
}

sort_resource_types = generic_sort(RESOURCE_TYPE_SORT_FUNCTIONS_BY_COLUMN)



# Copyright 2019-2020 SURF.
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from functools import partial
from http import HTTPStatus
from typing import Any, Dict, List, Optional
from uuid import UUID

import structlog
from celery import Celery, Task
from celery.app.control import Inspect
from celery.utils.log import get_task_logger
from kombu.serialization import registry

from orchestrator.api.error_handling import raise_status
from orchestrator.schemas.engine_settings import WorkerStatus
from orchestrator.services.processes import _get_process, _run_process_async, safe_logstep, thread_resume_process
from orchestrator.types import State
from orchestrator.utils.json import json_dumps, json_loads
from orchestrator.workflow import ProcessStat, Success, runwf
from orchestrator.workflows import get_workflow

logger = get_task_logger(__name__)

local_logger = structlog.get_logger(__name__)

_celery: Optional[Celery] = None

NEW_TASK = "tasks.new_task"
NEW_WORKFLOW = "tasks.new_workflow"
RESUME_TASK = "tasks.resume_task"
RESUME_WORKFLOW = "tasks.resume_workflow"


def get_celery_task(task_name: str) -> Task:
    if _celery:
        return _celery.signature(task_name)
    raise AssertionError("Celery has not been initialised yet")


def register_custom_serializer() -> None:
    # orchestrator specific serializer to correctly handle more complex classes
    registry.register("orchestrator-json", json_dumps, json_loads, "application/json", "utf-8")


def initialise_celery(celery: Celery) -> None:  # noqa: C901
    global _celery
    if _celery:
        raise AssertionError("You can only initialise Celery once")
    _celery = celery

    # Different routes/queues so we can assign them priorities
    celery.conf.task_routes = {
        NEW_TASK: {"queue": "new_tasks"},
        NEW_WORKFLOW: {"queue": "new_workflows"},
        RESUME_TASK: {"queue": "resume_tasks"},
        RESUME_WORKFLOW: {"queue": "resume_workflows"},
    }

    register_custom_serializer()

    def start_process(process_id: UUID, workflow_key: str, state: Dict[str, Any], user: str) -> Optional[UUID]:
        try:
            workflow = get_workflow(workflow_key)

            if not workflow:
                raise_status(HTTPStatus.NOT_FOUND, "Workflow does not exist")

            pstat = ProcessStat(
                process_id,
                workflow=workflow,
                state=Success(state),
                log=workflow.steps,
                current_user=user,
            )

            safe_logstep_with_func = partial(safe_logstep, broadcast_func=None)
            process_id = _run_process_async(pstat.process_id, lambda: runwf(pstat, safe_logstep_with_func))

        except Exception as exc:
            local_logger.error("Worker failed to execute workflow", process_id=process_id, details=str(exc))
            return None
        else:
            return process_id

    def resume_process(process_id: UUID, user_inputs: Optional[List[State]], user: str) -> Optional[UUID]:
        try:
            process = _get_process(process_id)
            process_id = thread_resume_process(process, user_inputs=user_inputs, user=user)
        except Exception as exc:
            local_logger.error("Worker failed to resume workflow", process_id=process_id, details=str(exc))
            return None
        else:
            return process_id

    celery_task = partial(celery.task, log=local_logger, serializer="orchestrator-json")

    @celery_task(name=NEW_TASK)  # type: ignore
    def new_task(process_id, workflow_key: str, state: Dict[str, Any], user: str) -> Optional[UUID]:
        local_logger.info("Start task", process_id=process_id, workflow_key=workflow_key)
        return start_process(process_id, workflow_key, state=state, user=user)

    @celery_task(name=NEW_WORKFLOW)  # type: ignore
    def new_workflow(process_id, workflow_key: str, state: Dict[str, Any], user: str) -> Optional[UUID]:
        local_logger.info("Start workflow", process_id=process_id, workflow_key=workflow_key)
        return start_process(process_id, workflow_key, state=state, user=user)

    @celery_task(name=RESUME_TASK)  # type: ignore
    def resume_task(process_id: UUID, user_inputs: Optional[List[State]], user: str) -> Optional[UUID]:
        local_logger.info("Resume task", process_id=process_id)
        return resume_process(process_id, user_inputs=user_inputs, user=user)

    @celery_task(name=RESUME_WORKFLOW)  # type: ignore
    def resume_workflow(process_id: UUID, user_inputs: Optional[List[State]], user: str) -> Optional[UUID]:
        local_logger.info("Resume workflow", process_id=process_id)
        return resume_process(process_id, user_inputs=user_inputs, user=user)


class CeleryJobWorkerStatus(WorkerStatus):
    def __init__(self) -> None:
        super().__init__(executor_type="celery")
        if not _celery:
            logger.error("Can't create CeleryJobStatistics. Celery is not initialised.")
            return

        inspection: Inspect = _celery.control.inspect()
        stats = inspection.stats()
        self.number_of_workers_online = len(stats)

        def sum_items(d: dict) -> int:
            return sum(len(lines) for _, lines in d.items()) if d else 0

        self.number_of_queued_jobs = sum_items(inspection.scheduled()) + sum_items(inspection.reserved())
        self.number_of_running_jobs = sum(len(tasks) for w, tasks in inspection.active().items())



# Copyright 2019-2020 SURF.
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from datetime import datetime
from http import HTTPStatus
from uuid import UUID

from fastapi.param_functions import Body
from fastapi.routing import APIRouter
from pytz import timezone

from orchestrator.api.error_handling import raise_status
from orchestrator.api.models import delete, save, update
from orchestrator.db import SubscriptionCustomerDescriptionTable
from orchestrator.schemas import SubscriptionDescriptionBaseSchema, SubscriptionDescriptionSchema
from orchestrator.utils.redis import delete_from_redis

router = APIRouter()


@router.post("/", response_model=None, status_code=HTTPStatus.NO_CONTENT)
def save_subscription_customer_description(data: SubscriptionDescriptionBaseSchema = Body(...)) -> None:
    save(SubscriptionCustomerDescriptionTable, data)
    delete_from_redis(data.subscription_id)


@router.put("/", response_model=None, status_code=HTTPStatus.NO_CONTENT)
def update_subscription_customer_descriptions(data: SubscriptionDescriptionSchema = Body(...)) -> None:
    if data.created_at is None:
        data.created_at = datetime.now(tz=timezone("UTC"))
    update(SubscriptionCustomerDescriptionTable, data)
    delete_from_redis(data.subscription_id)


@router.delete("/{_id}", response_model=None, status_code=HTTPStatus.NO_CONTENT)
def delete_subscription_customer_descriptions(_id: UUID) -> None:
    description = SubscriptionCustomerDescriptionTable.query.get(_id)
    delete(SubscriptionCustomerDescriptionTable, _id)
    delete_from_redis(description.subscription_id)


@router.get("/{_id}", response_model=SubscriptionDescriptionSchema)
def get_subscription_customer_descriptions(_id: UUID) -> str:
    description = SubscriptionCustomerDescriptionTable.query.get(_id)
    if description is None:
        raise_status(HTTPStatus.NOT_FOUND)
    return description


@router.get("/customer/{customer_id}/subscription/{subscription_id}", response_model=SubscriptionDescriptionSchema)
def get_subscription_customer_description_by_customer_subscription(customer_id: str, subscription_id: UUID) -> str:
    description = SubscriptionCustomerDescriptionTable.query.filter_by(
        customer_id=customer_id, subscription_id=subscription_id
    ).one_or_none()
    if description is None:
        raise_status(HTTPStatus.NOT_FOUND)
    return description



# Copyright 2019-2020 SURF.
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


import copy
import os
from http import HTTPStatus
from time import sleep
from typing import Any, TypedDict, Union
from uuid import UUID

import jsonref
import requests
import structlog
from more_itertools import first_true

from nwastdlib.url import URL
from orchestrator.types import State
from pydantic_forms.types import InputForm as LegacyInputForm


class JSONSubSchema(TypedDict, total=False):
    title: str
    type: str
    format: str
    pattern: str
    const: Any
    default: Any
    uniforms: dict
    enum: list
    allOf: list


class JSONSchema(JSONSubSchema, total=False):
    properties: JSONSubSchema
    items: JSONSubSchema


InputForm = Union[LegacyInputForm, JSONSchema]

try:
    # dotenv is not available during acceptance tests
    from dotenv import load_dotenv

    load_dotenv()
except ImportError:
    pass

logger = structlog.get_logger(__name__)


UUIDstr = str
ALLOWED_VLAN_RANGE = 4095


BASE_API_URL = URL(os.getenv("BASE_API_URL", "http://localhost:8080/api"))
TOKEN = os.getenv("TOKEN")
CWI = "2f47f65a-0911-e511-80d0-005056956c1a"


def resolve_all_of(input_field: JSONSchema) -> None:
    """Resolve allof in input field.

    Example:
        >>> field = {"allOf": [{"type": "string"}]}
        >>> resolve_all_of(field)
        >>> field
        {'allOf': [{'type': 'string'}], 'type': 'string'}
    """
    for object in input_field.get("allOf", []):
        input_field.update(object)


class Populator:
    """Base class for populating stuff, it will contain functionality to provide default input for all types."""

    def __init__(self, product_name: str) -> None:
        self.started = False
        self.done = False
        self.waiting_for_input = False
        self.last_state: dict = {}
        self.process_id = None
        self._product_name = product_name

        organisation = os.getenv("POPULATOR_ORGANISATION") or CWI
        contact_name = os.getenv("POPULATOR_CONTACT_NAME") or os.getenv("USER") or "unknown"
        contact_email = os.getenv("POPULATOR_CONTACT_EMAIL") or "a@b.nl"
        contact_phone = os.getenv("POPULATOR_CONTACT_PHONE") or ""

        # When a input_type is not found the populator will try to call a function self."resolve_"input_type()
        self.default_input_values = {
            "organisation": organisation,
            "guid": "6769d05f-3b11-e511-80d0-005056956c1a",
            "contact_persons": [{"email": contact_email, "name": contact_name, "phone": contact_phone}],
            "accept": "ACCEPTED",
            "ticket_id": "",
            "label": "",
        }
        self.custom_input_values: dict = {}
        self.add_default_values()
        self.session = requests.Session()
        if TOKEN:
            self.update_headers({"Authorization": "bearer " + TOKEN})

        self._product = self._get_product_by_name(self._product_name)
        self.create_workflow = first_true(
            self._product.get("workflows", []), {"name": "no_create_wf"}, lambda wf: wf["target"] == "CREATE"
        )["name"]
        if not hasattr(self, "log"):
            self.log = logger.bind()  # Silence type errors
            raise AssertionError("self.log must be set first")
        self.log = self.log.bind(product_name=self._product_name)
        self.log.info("Populator info", BASE_API_URL=BASE_API_URL)

        try:
            self.terminate_workflow = first_true(
                self._product.get("workflows", []), {"name": "no_terminate_wf"}, lambda wf: wf["target"] == "TERMINATE"
            )["name"]
        except TypeError:
            #  Terminate not yet implemented: Fall back to unbound logger to avoid logging problems
            logger.warning("Terminate workflow not implemented")
            self.terminate_workflow = None

        self.modify_workflows = list(  # noqa: C417
            map(lambda wf: wf["name"], filter(lambda wf: wf["target"] == "MODIFY", self._product.get("workflows", [])))
        )

    def update_headers(self, headers: dict) -> None:
        self.session.headers.update(headers)

    def add_default_values(self) -> None:
        """Add extra default value from classes that inherit me.

        This method can be overridden to add custom default
        values. Please ensure that you call super().add_default_values() in that case.
        """

    def _get_product_by_name(self, product_name: str) -> dict:
        """Fetch all active products.

        Returns: product dict

        """
        response = self.session.get(BASE_API_URL / "products/")
        if response.status_code == HTTPStatus.OK:
            for product in response.json():
                if product["name"] == product_name:
                    self.log.info("Resolved product.", product_id=product["product_id"])
                    return product
        self.log.error("Error product not found.", product=product_name, status_code=response.status_code)
        return {}

    def get_form_data(self, form: JSONSchema) -> dict:  # noqa: C901
        """Compiles a dict that can be used as the payload in api request that need human input.

        Note: if you want to populate a boolean field: you have to provide 'False' instead of 'None', as the
        latter is used to determine if one of the ways to populate the input_filed value was already successful.

        Args:
            form: a json schema form definition.

        Returns: a dict that can be used as payload.

        """

        self.log.info("Get form data for input fields.", form=form)
        data = {}

        # Resolve refs
        form = jsonref.JsonRef.replace_refs(form)

        field_name: str
        input_field: JSONSchema
        for field_name, input_field in form["properties"].items():  # type: ignore
            resolve_all_of(input_field)

            log = self.log.bind(field_name=field_name)
            # Todo: make toggle?
            # log = self.log.bind(input_field=input_field)

            # Read only always has a value that should be returned as is
            if "const" in input_field:
                value = input_field.get("const")
            else:
                value = None

            if value is None:
                value = self.custom_input_values.get(field_name)

            if value is None:
                value = self.custom_input_values.get(input_field.get("format"))

            if value is None:
                value = self.custom_input_values.get(input_field["type"])

            # Before resolving we check if we cannot use the default/current
            if value is None:
                value = input_field.get("default")

            if value is None and field_name:
                value = self.default_input_values.get(field_name)

            if value is None:
                value = self.default_input_values.get(input_field.get("format", ""))

            if value is None:
                value = self.default_input_values.get(input_field["type"])

            if value is None:
                # try to call a function based on the field_name
                log.info("Trying to resolve input field with custom function based on name.")
                try:
                    func_name = "resolve_" + field_name.replace(".", "_")
                    log = log.bind(func_name=func_name)
                    log.info("Calling custom function.")
                    value = getattr(self, func_name)(input_field)
                except AttributeError:
                    log.warning("Unable to resolve custom function based on name.")
                    value = None

            if value is None:
                # try to call a function based on the input_field["format"]
                log.info("Trying to resolve input field with custom function based on format.")
                try:
                    func_name = f"resolve_{input_field.get('format')}"
                    log = log.bind(func_name=func_name)
                    log.info("Calling custom function.")
                    value = getattr(self, func_name)(input_field)
                except AttributeError:
                    log.warning("Unable to resolve custom function based on format.")
                    value = None

            if value is None:
                # try to call a function based on the input_field["type"]
                log.info("Trying to resolve input field with custom function based on type.")
                try:
                    func_name = f"resolve_{input_field['type']}"
                    log = log.bind(func_name=func_name)
                    log.info("Calling custom function.")
                    value = getattr(self, func_name)(input_field)
                except AttributeError:
                    log.warning("Unable to resolve custom function based on type.")
                    value = None

            # If enum just pick the first or leave empty if there are no options to select
            if value is None and "enum" in input_field and input_field["enum"]:
                value = input_field["enum"][0]

            if value is None and input_field.get("format") == "divider":
                # Ignore divider elements
                value = ""

            log.debug("Resolved input_field.", value=value)
            data[field_name] = value
        return data

    def _start_workflow(self, workflow_name: str, **kwargs: Any) -> UUIDstr:
        """Start a workflow.

        Args:
            workflow_name: workflow name
            kwargs: The kwargs

        Returns: the process_id of the workflow process

        """
        self.log = self.log.bind(process_id=None, workflow=workflow_name)
        self.log.info("Starting workflow")

        self.custom_input_values.update(kwargs)

        response = self.provide_user_input("POST", BASE_API_URL / "processes" / workflow_name)

        response_json = response.json()
        if "id" not in response_json:
            raise Exception(f"Starting workflow {workflow_name} failed: {response_json}")

        self.process_id = process_id = response_json["id"]
        self.log = self.log.bind(process_id=self.process_id)
        self.started = True
        return process_id

    def start_create_workflow(self, **kwargs: Any) -> UUIDstr:
        """Start a create workflow.

        Args:
            kwargs: values to be used as form input

        Returns: the process_id of the workflow process

        """
        self.log = self.log.bind(subscription_id=None)
        self.log.info("Started create workflow")
        product_id = self._product.get("product_id")
        return self._start_workflow(self.create_workflow, product=product_id, **kwargs)

    def start_modify_workflow(
        self, workflow_name: str, subscription_id: Union[UUIDstr, UUID], **kwargs: Any
    ) -> UUIDstr:
        """Start a modify workflow for the provided name and subscription_id.

        Args:
            workflow_name: workflow name
            subscription_id: uuid of the subscription you want to modify
            kwargs: values to be used as form input

        Returns: the process_id of the workflow process

        """
        subscription_id = str(subscription_id)
        self.log = self.log.bind(subscription_id=subscription_id)
        self.log.info("Started modify workflow")
        return self._start_workflow(workflow_name, subscription_id=subscription_id, **kwargs)

    def start_verify_workflow(self, workflow_name: str, subscription_id: Union[UUIDstr, UUID]) -> UUIDstr:
        subscription_id = str(subscription_id)
        self.log = self.log.bind(subscription_id=subscription_id)
        self.log.info("Started verify workflow")
        return self._start_workflow(workflow_name, subscription_id=subscription_id)

    def start_terminate_workflow(self, subscription_id: Union[UUIDstr, UUID], **kwargs: Any) -> UUIDstr:
        """Start a terminate workflow for the provided subscription_id.

        Args:
            subscription_id: uuid of the subscription you want to terminate
            kwargs: values to be used as form input

        Returns: the process_id of the workflow process

        """
        subscription_id = str(subscription_id)
        self.log = self.log.bind(process_id=None, subscription_id=subscription_id)
        self.log.info("Starting terminate workflow")
        return self._start_workflow(self.terminate_workflow, subscription_id=subscription_id, **kwargs)

    def human_input_needed(self) -> bool:
        """Check whether the workflow process needs human input.

        Returns: True or False

        """
        response = self.session.get(BASE_API_URL / "processes" / self.process_id)
        if response.status_code == HTTPStatus.OK:
            self.last_state = response.json()
        else:
            self.log.error("Cowardly quitting due to response code.", status_code=response.status_code)
            raise Exception("Request failed")
        status = self.last_state["status"] if "status" in self.last_state else self.last_state["last_status"]
        if status == "completed":
            self.log.info("Process is complete.")
            self.done = True
            return False

        if status == "suspended":
            return True

        if status in ("created", "running", "resumed"):
            return False

        if status in ("failed", "waiting"):
            if self.retries < 1:
                self.retries += 1
                return True

            self.log.error("Cowardly quitting due to failed step.", reason=self.last_state["failed_reason"])
            raise Exception(f"Step failed: {self.last_state['failed_reason']}")
        self.log.error("Cowardly quitting due to unknown status", status=self.last_state["status"])
        raise Exception(f"Unknown status: {status}")

    def get_current_form(self) -> Union[JSONSchema, None]:
        self.log.info("Current form.", form=self.last_state.get("form"))
        return copy.deepcopy(self.last_state.get("form"))

    def provide_user_input(self, method: str, url: URL, form: Union[JSONSchema, None] = None) -> requests.Response:
        """Provide input for steps that normally require a user."""
        self.log.info("Providing user input.")

        user_inputs: list[State] = [self.get_form_data(form)] if form else []
        while True:
            self.log.info("Submitting user input", data=user_inputs)
            response = self.session.request(method, url, json=user_inputs)
            if response.status_code != HTTPStatus.NOT_EXTENDED:
                return response
            input_fields = response.json()["form"]
            user_inputs.append(self.get_form_data(input_fields))

    def reset(self) -> None:
        """Reset internal state."""
        self.log = self.log.bind(subscription_id=None, process_id=None).unbind("subscription_id", "process_id")
        self.started = False
        self.done = False
        self.waiting_for_input = False
        self.last_state = {}
        self.process_id = None

        # ensure dynamic stuff is also reset
        self.add_default_values()

    def run(self, **kwargs: Any) -> UUIDstr:
        """Responsible for auto-completing the workflow after the process has been started.

        Returns: subscription_id of the created subscription

        """
        self.custom_input_values.update(kwargs)
        self.retries = 0

        while self.started and not self.done:
            self.log.info("Sleeping.")
            sleep(1.0)
            if self.human_input_needed():
                form = self.get_current_form()

                response = self.provide_user_input("PUT", BASE_API_URL / "processes" / self.process_id / "resume", form)
                if response.status_code == HTTPStatus.NO_CONTENT:
                    self.log.info("Resumed process.")
                else:
                    self.log.error("Resume process failed.", status_code=response.status_code)
                    raise Exception("Resume process failed")

        subscription_id = self.last_state["current_state"]["subscription_id"]
        self.log.info("Finished workflow", subscription_id=subscription_id)

        self.reset()

        return subscription_id


class TerminatePopulator(Populator):
    def __init__(self) -> None:
        self.log = logger.bind()
        super().__init__("IP_PREFIX")  # Dummy product name because we don't care for terminates



# Copyright 2022-2023 SURF.
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from collections.abc import Awaitable
from ipaddress import IPv4Address, IPv4Interface, IPv6Address, IPv6Interface

# Map some Orchestrator types to scalars
from typing import Any, Callable, List, NewType, Tuple, Union

import strawberry
from graphql import GraphQLError
from starlette.requests import Request
from strawberry.custom_scalar import ScalarDefinition, ScalarWrapper
from strawberry.scalars import JSON
from strawberry.types import Info
from strawberry.types.info import RootValueType

from oauth2_lib.fastapi import OIDCUserModel
from oauth2_lib.strawberry import OauthContext
from orchestrator.db.filters import Filter
from orchestrator.db.sorting import Sort, SortOrder
from orchestrator.utils.vlans import VlanRanges


def serialize_to_string(value: Any) -> str:
    return str(value)


def serialize_vlan(vlan: VlanRanges) -> List[Tuple[int, int]]:
    return vlan.to_list_of_tuples()


class OrchestratorContext(OauthContext):
    def __init__(
        self,
        get_current_user: Callable[[Request], Awaitable[OIDCUserModel]],
        get_opa_decision: Callable[[str, OIDCUserModel], Awaitable[Union[bool, None]]],
    ):
        self.errors: list[GraphQLError] = []
        super().__init__(get_current_user, get_opa_decision)


OrchestratorInfo = Info[OrchestratorContext, RootValueType]


@strawberry.experimental.pydantic.input(model=Sort)
class GraphqlSort:
    field: str = strawberry.field(description="Field to sort on")
    order: SortOrder = strawberry.field(default=SortOrder.ASC, description="Sort order (ASC or DESC")


@strawberry.experimental.pydantic.input(model=Filter)
class GraphqlFilter:
    field: str = strawberry.field(description="Field to filter on")
    value: str = strawberry.field(description="Value to sort the field on")


VlanRangesType = strawberry.scalar(
    NewType("VlanRangesType", str),
    description="Represent the Orchestrator VlanRanges data type",
    serialize=serialize_vlan,
    parse_value=lambda v: v,
)

IPv4AddressType = strawberry.scalar(
    NewType("IPv4AddressType", str),
    description="Represent the Orchestrator IPv4Address data type",
    serialize=serialize_to_string,
    parse_value=lambda v: v,
)

IPv6AddressType = strawberry.scalar(
    NewType("IPv6AddressType", str),
    description="Represent the Orchestrator IPv6Address data type",
    serialize=serialize_to_string,
    parse_value=lambda v: v,
)

IPv4InterfaceType = strawberry.scalar(
    NewType("IPv4InterfaceType", str),
    description="Represent the Orchestrator IPv4Interface data type",
    serialize=serialize_to_string,
    parse_value=lambda v: v,
)

IPv6InterfaceType = strawberry.scalar(
    NewType("IPv6InterfaceType", str),
    description="Represent the Orchestrator IPv6Interface data type",
    serialize=serialize_to_string,
    parse_value=lambda v: v,
)

SCALAR_OVERRIDES: dict[object, Union[Any, ScalarWrapper, ScalarDefinition]] = {
    dict: JSON,
    VlanRanges: VlanRangesType,
    IPv4Address: IPv4AddressType,
    IPv6Address: IPv6AddressType,
    IPv4Interface: IPv4InterfaceType,
    IPv6Interface: IPv6InterfaceType,
}



# Copyright 2019-2020 SURF.
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from copy import deepcopy
from typing import Any, Optional, Set, Union
from uuid import UUID

import structlog
from pydantic import ValidationError

from orchestrator.db import db
from orchestrator.db.models import ProcessSubscriptionTable
from orchestrator.domain.base import ProductBlockModel, SubscriptionModel
from orchestrator.services.subscriptions import build_extended_domain_model, get_subscription
from orchestrator.targets import Target
from orchestrator.types import State, SubscriptionLifecycle, UUIDstr
from orchestrator.utils.json import to_serializable
from orchestrator.utils.redis import delete_from_redis, to_redis
from orchestrator.workflow import Step, step

logger = structlog.get_logger(__name__)


@step("Unlock subscription")
def resync(subscription: SubscriptionModel) -> State:
    """Transition a subscription to in sync."""
    subscription.insync = True
    return {"subscription": subscription}


@step("Lock subscription")
def unsync(subscription_id: UUIDstr, __old_subscriptions__: Optional[dict] = None) -> State:
    """Transition a subscription to out of sync.

    This step will also create a backup of the current subscription details in the state with the key
    `__old_subscriptions__`

    Note:  This step will NOT overwrite any existing data in the state for the `__old_subscriptions__` key. Some
    workflows already change subscription details in the initial form step. This is not a best practice but
    it can be handy/needed for some scenarios. To ensure that you get a correct backup these forms need to create
    the backup in the initial form step and save it to `__old_subscriptions__`.

    An example, showing the creation of a backup for 2 subscriptions during the initial form step :

    ```
    subscription = YOUR_DOMAIN_MODEL.from_subscription(subscription_id)
    subscription_backup = {}
    subscription_backup[subscription.subscription_id] = deepcopy(subscription.dict())
    subscription_backup[second_subscription.subscription_id] = deepcopy(second_subscription.dict())
    return {..., "__old_subscriptions__": subscription_backup }
    ```

    It's also possible to make the backup for a subscription that doesn't have a domain model:

    ```
    from orchestrator.utils.json import to_serializable
    subscription = subscriptions.get_subscription(subscription_id)
    subscription_backup = {subscription_id: to_serializable(subscription)}
    Do your changes here ...
    return {..., "subscription": to_serializable(subscription), "__old_subscriptions__": subscription_backup }
    ```

    """
    try:
        subscription = SubscriptionModel.from_subscription(subscription_id)
    except ValidationError:
        subscription = get_subscription(subscription_id)  # type: ignore

    # Handle backup if needed
    if __old_subscriptions__ and __old_subscriptions__.get(subscription_id):
        logger.info(
            "Skipping backup of subscription details because it already exists in the state",
            subscription_id=str(subscription_id),
        )
        subscription_backup = __old_subscriptions__
    else:
        logger.info("Creating backup of subscription details in the state", subscription_id=str(subscription_id))
        subscription_backup = __old_subscriptions__ or {}
        if isinstance(subscription, SubscriptionModel):
            subscription_backup[str(subscription_id)] = deepcopy(subscription.dict())
        else:
            subscription_backup[str(subscription_id)] = to_serializable(subscription)  # type: ignore

    # Handle transition
    if not subscription.insync:
        raise ValueError("Subscription is already out of sync, cannot continue!")
    subscription.insync = False

    return {"subscription": subscription, "__old_subscriptions__": subscription_backup}


@step("Lock subscription")
def unsync_unchecked(subscription_id: UUIDstr) -> State:
    """Use for validation workflows that need to run if the subscription is out of sync."""
    subscription = SubscriptionModel.from_subscription(subscription_id)
    subscription.insync = False
    return {"subscription": subscription}


def store_process_subscription_relationship(
    process_id: UUIDstr, subscription_id: UUIDstr, workflow_target: str
) -> ProcessSubscriptionTable:
    process_subscription = ProcessSubscriptionTable(
        process_id=process_id, subscription_id=subscription_id, workflow_target=workflow_target
    )
    db.session.add(process_subscription)
    return process_subscription


def store_process_subscription(workflow_target: Target) -> Step:
    @step("Create Process Subscription relation")
    def _store_process_subscription(process_id: UUIDstr, subscription_id: UUIDstr) -> None:
        store_process_subscription_relationship(process_id, subscription_id, workflow_target)

    return _store_process_subscription


def set_status(status: SubscriptionLifecycle) -> Step:
    @step(f"Set subscription to '{status}'")
    def _set_status(subscription: SubscriptionModel) -> State:
        """Set subscription to status."""
        subscription = SubscriptionModel.from_other_lifecycle(subscription, status)
        return {"subscription": subscription}

    _set_status.__doc__ = f"Set subscription to '{status}'."
    return _set_status


@step("Remove domain model from cache")
def remove_domain_model_from_cache(
    workflow_name: str, subscription: Optional[SubscriptionModel] = None, subscription_id: Optional[UUID] = None
) -> State:
    """Remove the domain model from the cache if it exists.

    Args:
        workflow_name: The workflow name
        subscription: Subscription Model
        subscription_id: The subscription id

    Returns:
        State

    """

    if not (subscription or subscription_id):
        logger.warning("No subscription found in this workflow", workflow_name=workflow_name)
        return {"deleted_subscription_id": None}
    if subscription:
        delete_from_redis(subscription.subscription_id)
    elif subscription_id:
        delete_from_redis(subscription_id)

    return {"deleted_subscription_id": subscription_id or subscription.subscription_id}  # type: ignore[union-attr]


@step("Cache Subscription and related subscriptions")
def cache_domain_models(workflow_name: str, subscription: Optional[SubscriptionModel] = None) -> State:  # noqa: C901
    """Attempt to cache all Subscriptions once they have been touched once.

    Args:
        workflow_name: The Workflow Name
        subscription:  The Subscription if it exists.

    Returns:
        Returns State.

    """
    cached_subscription_ids: Set[UUID] = set()
    if not subscription:
        logger.warning("No subscription found in this workflow", workflow_name=workflow_name)
        return {"cached_subscription_ids": cached_subscription_ids}

    def _cache_other_subscriptions(product_block: ProductBlockModel) -> None:
        for field in product_block.__fields__:
            # subscription_instance is a ProductBlockModel or an arbitrary type
            subscription_instance: Union[ProductBlockModel, Any] = getattr(product_block, field)
            # If subscription_instance is a list, we need to step into it and loop.
            if isinstance(subscription_instance, list):
                for item in subscription_instance:
                    if isinstance(item, ProductBlockModel):
                        _cache_other_subscriptions(item)

            # If subscription_instance is a ProductBlockModel check the owner_subscription_id to decide the cache
            elif isinstance(subscription_instance, ProductBlockModel):
                _cache_other_subscriptions(subscription_instance)
                if not subscription_instance.owner_subscription_id == subscription.subscription_id:
                    cached_subscription_ids.add(subscription_instance.owner_subscription_id)

    for field in subscription.__fields__:
        # There always is a single Root Product Block, it cannot be a list, so no need to check.
        instance: Union[ProductBlockModel, Any] = getattr(subscription, field)
        if isinstance(instance, ProductBlockModel):
            _cache_other_subscriptions(instance)

    # Cache all the sub subscriptions
    for subscription_id in cached_subscription_ids:
        subscription_model = SubscriptionModel.from_subscription(subscription_id)
        to_redis(build_extended_domain_model(subscription_model))

    # Cache the main subscription
    to_redis(build_extended_domain_model(subscription))
    cached_subscription_ids.add(subscription.subscription_id)

    return {"cached_subscription_ids": cached_subscription_ids}



# Copyright 2019-2020 SURF.
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from types import new_class
from typing import Any, Dict, Generator, List, Optional, Sequence, Type
from uuid import UUID

import structlog
from pydantic.errors import EnumMemberError
from pydantic.validators import uuid_validator

from orchestrator.services import products
from pydantic_forms.types import strEnum
from pydantic_forms.validators import (  # noqa: F401
    Accept,
    Choice,
    ChoiceList,
    ContactPerson,
    ContactPersonList,
    ContactPersonName,
    DisplaySubscription,
    Divider,
    Label,
    ListOfOne,
    ListOfTwo,
    LongText,
    MigrationSummary,
    OrganisationId,
    Timestamp,
    UniqueConstrainedList,
    choice_list,
    contact_person_list,
    migration_summary,
    remove_empty_items,
    timestamp,
    unique_conlist,
)

logger = structlog.get_logger(__name__)

__all__ = [
    "Accept",
    "Choice",
    "ChoiceList",
    "ContactPerson",
    "ContactPersonName",
    "ContactPersonList",
    "DisplaySubscription",
    "Divider",
    "Label",
    "ListOfOne",
    "ListOfTwo",
    "LongText",
    "ProductIdError",
    "ProductId",
    "MigrationSummary",
    "OrganisationId",
    "Timestamp",
    "UniqueConstrainedList",
    "choice_list",
    "contact_person_list",
    "migration_summary",
    "product_id",
    "remove_empty_items",
    "strEnum",
    "timestamp",
    "unique_conlist",
]


class ProductIdError(EnumMemberError):
    code = "product_id"
    enum_values: Sequence[Any]  # Required kwarg

    def __str__(self) -> str:
        permitted = ", ".join(repr(v) for v in self.enum_values)
        return f"value is not a valid enumeration member; permitted: {permitted}"


class ProductId(UUID):
    products: Optional[List[UUID]] = None

    @classmethod
    def __modify_schema__(cls, field_schema: Dict[str, Any]) -> None:
        kwargs = {"uniforms": {"productIds": cls.products}} if cls.products else {}
        field_schema.update(format="productId", **kwargs)

    @classmethod
    def __get_validators__(cls) -> Generator:
        yield uuid_validator
        yield cls.is_product
        if cls.products:
            yield cls.in_products

    @classmethod
    def is_product(cls, v: UUID) -> UUID:
        product = products.get_product_by_id(v)
        if product is None:
            raise ValueError("Product not found")

        return v

    @classmethod
    def in_products(cls, v: UUID) -> UUID:
        if cls.products and v not in cls.products:
            raise ProductIdError(enum_values=list(map(str, cls.products)))

        return v


def product_id(products: Optional[List[UUID]] = None) -> Type[ProductId]:
    namespace = {"products": products}
    return new_class("ProductIdSpecific", (ProductId,), {}, lambda ns: ns.update(namespace))



# Copyright 2019-2020 SURF.
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from authlib.integrations.starlette_client import OAuth

from nwastdlib.url import URL
from oauth2_lib.fastapi import HTTPX_SSL_CONTEXT, OIDCUser, opa_decision, opa_graphql_decision
from oauth2_lib.settings import oauth2lib_settings
from orchestrator.settings import oauth2_settings

oauth_client_credentials = OAuth()

well_known_endpoint = URL(oauth2_settings.OIDC_CONF_WELL_KNOWN_URL)

oauth_client_credentials.register(
    "connext",
    server_metadata_url=well_known_endpoint / ".well-known" / "openid-configuration",
    client_id=oauth2_settings.OAUTH2_RESOURCE_SERVER_ID,
    client_secret=oauth2_settings.OAUTH2_RESOURCE_SERVER_SECRET,
    request_token_params={"grant_type": "client_credentials"},
    client_kwargs={"verify": HTTPX_SSL_CONTEXT},
)

oidc_user = OIDCUser(
    oauth2_settings.OIDC_CONF_WELL_KNOWN_URL,
    oauth2_settings.OAUTH2_RESOURCE_SERVER_ID,
    oauth2_settings.OAUTH2_RESOURCE_SERVER_SECRET,
    enabled=oauth2lib_settings.OAUTH2_ACTIVE,
)

opa_security_default = opa_decision(oauth2_settings.OPA_URL, oidc_user, enabled=oauth2lib_settings.OAUTH2_ACTIVE)


opa_security_graphql = opa_graphql_decision(
    oauth2_settings.OPA_URL, oidc_user, enabled=oauth2lib_settings.OAUTH2_ACTIVE
)


def get_oidc_user() -> OIDCUser:
    return oidc_user


def get_opa_security_graphql():  # type: ignore
    return opa_security_graphql



# Copyright 2019-2020 SURF.
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import Union

import structlog
from pydantic.utils import to_lower_camel

from orchestrator.db import ProductTable, SubscriptionTable
from orchestrator.db.filters import Filter
from orchestrator.db.filters.subscription import filter_subscriptions
from orchestrator.db.range import apply_range_to_query
from orchestrator.db.sorting import Sort, sort_subscriptions
from orchestrator.domain.base import SubscriptionModel
from orchestrator.graphql.pagination import Connection
from orchestrator.graphql.schemas.product import ProductModelGraphql
from orchestrator.graphql.schemas.subscription import Subscription, SubscriptionInterface
from orchestrator.graphql.types import GraphqlFilter, GraphqlSort, OrchestratorInfo
from orchestrator.graphql.utils.create_resolver_error_handler import create_resolver_error_handler
from orchestrator.graphql.utils.is_query_detailed import is_query_detailed
from orchestrator.graphql.utils.to_graphql_result_page import to_graphql_result_page
from orchestrator.types import SubscriptionLifecycle

logger = structlog.get_logger(__name__)
# Note: we can make this more fancy by adding metadata to the field annotation that indicates if a resolver
# needs subscription details or not, and use that information here. Left as an exercise for the reader.
base_sub_props = tuple(
    [to_lower_camel(key) for key in SubscriptionInterface.__annotations__]
    + [to_lower_camel(key) for key in ProductModelGraphql.__annotations__]
    + ["__typename"]
)

_is_subscription_detailed = is_query_detailed(base_sub_props)


def get_subscription_details(subscription: SubscriptionTable) -> SubscriptionInterface:
    from orchestrator.graphql.autoregistration import graphql_subscription_name
    from orchestrator.graphql.schema import GRAPHQL_MODELS

    subscription_details = SubscriptionModel.from_subscription(subscription.subscription_id)
    base_model = subscription_details.__base_type__ if subscription_details.__base_type__ else subscription_details
    subscription_details = base_model.from_other_lifecycle(  # type: ignore
        subscription_details, SubscriptionLifecycle.INITIAL, skip_validation=True
    )
    strawberry_type = GRAPHQL_MODELS[graphql_subscription_name(base_model.__name__)]  # type: ignore
    return strawberry_type.from_pydantic(subscription_details)


async def resolve_subscriptions(
    info: OrchestratorInfo,
    filter_by: Union[list[GraphqlFilter], None] = None,
    sort_by: Union[list[GraphqlSort], None] = None,
    first: int = 10,
    after: int = 0,
) -> Connection[SubscriptionInterface]:
    _error_handler = create_resolver_error_handler(info)

    pydantic_filter_by: list[Filter] = [item.to_pydantic() for item in filter_by] if filter_by else []
    pydantic_sort_by: list[Sort] = [item.to_pydantic() for item in sort_by] if sort_by else []
    logger.debug("resolve_subscription() called", range=[after, after + first], sort=sort_by, filter=pydantic_filter_by)

    query = SubscriptionTable.query.join(ProductTable)

    query = filter_subscriptions(query, pydantic_filter_by, _error_handler)
    query = sort_subscriptions(query, pydantic_sort_by, _error_handler)
    total = query.count()
    query = apply_range_to_query(query, after, first)

    subscriptions = query.all()
    if _is_subscription_detailed(info):
        graphql_subscriptions = [get_subscription_details(p) for p in subscriptions]
    else:
        graphql_subscriptions = [Subscription.from_pydantic(p) for p in subscriptions]
    return to_graphql_result_page(graphql_subscriptions, first, after, total)



# Copyright 2019-2020 SURF.
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from enum import Enum  # noqa: F401 (doctest)
from http import HTTPStatus
from typing import Any, Callable, Dict, List, Literal, Optional, Tuple, Union
from uuid import UUID

import strawberry
from pydantic.typing import get_args, get_origin, is_union

# TODO: eventually enforce code migration for downstream users to import
# these types from pydantic_forms themselves
from pydantic_forms.types import (
    JSON,
    AcceptData,
    AcceptItemType,
    FormGenerator,
    FormGeneratorAsync,
    InputForm,
    InputFormGenerator,
    InputStepFunc,
    SimpleInputFormGenerator,
    State,
    StateInputFormGenerator,
    StateInputFormGeneratorAsync,
    StateInputStepFunc,
    StateSimpleInputFormGenerator,
    SubscriptionMapping,
    SummaryData,
    UUIDstr,
    strEnum,
)

__all__ = [
    "JSON",
    "BroadcastFunc",
    "AcceptData",
    "AcceptItemType",
    "ErrorDict",
    "ErrorState",
    "FormGenerator",
    "FormGeneratorAsync",
    "InputForm",
    "InputFormGenerator",
    "InputStepFunc",
    "SimpleInputFormGenerator",
    "State",
    "StateInputFormGenerator",
    "StateInputFormGeneratorAsync",
    "StateInputStepFunc",
    "StateSimpleInputFormGenerator",
    "StateStepFunc",
    "StepFunc",
    "SubscriptionLifecycle",
    "SubscriptionMapping",
    "SummaryData",
    "UUIDstr",
    "is_list_type",
    "is_of_type",
    "is_optional_type",
    "is_union_type",
    "get_possible_product_block_types",
    "strEnum",
]

# ErrorState is either a string containing an error message, a catched Exception or a tuple containing a message and
# a HTTP status code
ErrorState = Union[str, Exception, Tuple[str, Union[int, HTTPStatus]]]
# An ErrorDict should have the following keys:
# error: str  # A message describing the error
# class: str[Optional]  # The exception class name (type)
# status_code: Optional[int]  # HTTP Status code (optional)
# traceback: Optional[str]  # A python traceback as a string formatted by nwastdlib.ex.show_ex
ErrorDict = Dict[str, Union[str, int, List[Dict[str, Any]], InputForm, None]]
StateStepFunc = Callable[[State], State]
StepFunc = Callable[..., Optional[State]]
BroadcastFunc = Callable[[UUID, Dict], None]


@strawberry.enum
class SubscriptionLifecycle(strEnum):
    INITIAL = "initial"
    ACTIVE = "active"
    MIGRATING = "migrating"
    DISABLED = "disabled"
    TERMINATED = "terminated"
    PROVISIONING = "provisioning"


# TODO #1321: old code that protected against unsafe changes in subs
# The key is the Parent subscription life cycle status. The keys are lists of safe transitions for child subscriptions.
SAFE_USED_BY_TRANSITIONS_FOR_STATUS = {
    SubscriptionLifecycle.INITIAL: [
        SubscriptionLifecycle.INITIAL,
    ],
    SubscriptionLifecycle.ACTIVE: [
        SubscriptionLifecycle.INITIAL,
        SubscriptionLifecycle.PROVISIONING,
        SubscriptionLifecycle.MIGRATING,
        SubscriptionLifecycle.ACTIVE,
        SubscriptionLifecycle.TERMINATED,
        SubscriptionLifecycle.DISABLED,
    ],
    SubscriptionLifecycle.MIGRATING: [
        SubscriptionLifecycle.INITIAL,
        SubscriptionLifecycle.MIGRATING,
        SubscriptionLifecycle.TERMINATED,
    ],
    SubscriptionLifecycle.PROVISIONING: [
        SubscriptionLifecycle.INITIAL,
        SubscriptionLifecycle.PROVISIONING,
        SubscriptionLifecycle.ACTIVE,
        SubscriptionLifecycle.TERMINATED,
    ],
    SubscriptionLifecycle.TERMINATED: [SubscriptionLifecycle.INITIAL, SubscriptionLifecycle.TERMINATED],
    SubscriptionLifecycle.DISABLED: [
        SubscriptionLifecycle.INITIAL,
        SubscriptionLifecycle.DISABLED,
        SubscriptionLifecycle.TERMINATED,
    ],
}


def is_of_type(t: Any, test_type: Any) -> bool:
    """Check if annotation type is valid for type.

    >>> is_of_type(list, list)
    True
    >>> is_of_type(List[int], List[int])
    True
    >>> is_of_type(strEnum, str)
    True
    >>> is_of_type(strEnum, Enum)
    True
    >>> is_of_type(int, str)
    False
    """

    if is_union_type(test_type):
        return any(get_origin(t) is get_origin(arg) for arg in get_args(test_type))

    if (
        get_origin(t)
        and get_origin(test_type)
        and get_origin(t) is get_origin(test_type)
        and get_args(t) == get_args(test_type)
    ):
        return True

    if test_type is t:
        # Test type is a typing type instance and matches
        return True

    # Workaround for the fact that you can't call issubclass on typing types
    try:
        return issubclass(t, test_type)
    except TypeError:
        return False


# TODO #1330: Fix comparison of is_of_type(Union[str, int], Union[int, str]), it returns False but it should be True
def is_list_type(t: Any, test_type: Optional[type] = None) -> bool:
    """Check if `t` is list type.

    And optionally check if the list items are of `test_type`

    >>> is_list_type(List[int])
    True
    >>> is_list_type(Optional[List[int]])
    True
    >>> is_list_type(Optional[List[int]], int)
    True
    >>> is_list_type(Optional[List[int]], str)
    False
    >>> is_list_type(Optional[int])
    False
    >>> is_list_type(List[Tuple[int, int]])
    True
    >>> is_list_type(List[Tuple[int, int]], int)
    False
    >>> is_list_type(List[Tuple[int, int]], Tuple[int, int])
    True
    >>> is_list_type(List[strEnum], Enum)
    True
    >>> is_list_type(int)
    False
    >>> is_list_type(Literal[1,2,3])
    False
    >>> is_list_type(List[Union[str, int]])
    True
    >>> is_list_type(List[Union[str, int]], Union[str, int])
    False
    >>> is_list_type(List[Union[str, int]], str)
    True
    >>> is_list_type(List[Union[str, int]], int)
    False
    >>> is_list_type(List[Union[str, int]], Union[int, int])
    False
    """
    if get_origin(t):
        if is_optional_type(t) or is_union_type(t):
            for arg in get_args(t):
                if is_list_type(arg, test_type):
                    return True
        elif get_origin(t) == Literal:  # type:ignore
            return False  # Literal cannot contain lists see pep 586
        elif issubclass(get_origin(t), list):  # type: ignore
            if test_type and get_args(t):
                first_arg = get_args(t)[0]
                # To support a list with union of multiple product blocks.
                if is_union_type(first_arg) and get_args(first_arg) and not is_union_type(test_type):
                    first_arg = get_args(first_arg)[0]
                return is_of_type(first_arg, test_type)
            return True

    return False


def is_optional_type(t: Any, test_type: Optional[type] = None) -> bool:
    """Check if `t` is optional type (Union[None, ...]).

    And optionally check if T is of `test_type`

    >>> is_optional_type(Optional[int])
    True
    >>> is_optional_type(Union[None, int])
    True
    >>> is_optional_type(Union[int, str, None])
    True
    >>> is_optional_type(Optional[int], int)
    True
    >>> is_optional_type(Optional[int], str)
    False
    >>> is_optional_type(Optional[State], int)
    False
    >>> is_optional_type(Optional[State], State)
    True
    """
    if is_union(get_origin(t)) and None.__class__ in get_args(t):
        for arg in get_args(t):
            if arg is None.__class__:
                continue

            return not test_type or is_of_type(arg, test_type)
    return False


def is_union_type(t: Any, test_type: Optional[type] = None) -> bool:
    """Check if `t` is union type (Union[Type, AnotherType]).

    Optionally check if T is of `test_type` We cannot check for literal Nones.

    >>> is_union_type(Union[int, str])
    True
    >>> is_union_type(Union[int, str], str)
    True
    >>> is_union_type(Union[int, str], Union[int, str])
    True
    >>> is_union_type(Union[int, None])
    True
    >>> is_union_type(int)
    False
    """
    if not is_union(get_origin(t)):
        return False
    if not test_type:
        return True

    if is_of_type(t, test_type):
        return True
    for arg in get_args(t):
        result = is_of_type(arg, test_type)
        if result:
            return result
    return False


def get_possible_product_block_types(list_field_type: Any) -> dict:
    possible_product_block_types = {}
    if is_union_type(list_field_type):
        for list_item_field_type in get_args(list_field_type):
            if (
                not isinstance(None, list_item_field_type)
                and list_item_field_type.name not in possible_product_block_types
            ):
                possible_product_block_types[list_item_field_type.name] = list_item_field_type
    else:
        possible_product_block_types = {list_field_type.name: list_field_type}
    return possible_product_block_types



from typing import Any

from orchestrator.graphql.pagination import Connection, PageInfo


def to_graphql_result_page(items: list[Any], first: int, after: int, total: int) -> Connection:
    has_next_page = len(items) > first

    page_items = items[:first]
    page_items_length = len(page_items)
    start_cursor = after if page_items_length else None
    end_cursor = after + page_items_length - 1

    return Connection(
        page=page_items,
        page_info=PageInfo(
            has_previous_page=bool(after),
            has_next_page=has_next_page,
            start_cursor=start_cursor,
            end_cursor=end_cursor,
            total_items=total if total else None,
        ),
    )



from orchestrator.db.filters.filters import (
    CallableErrorHandler,
    Filter,
    generic_apply_filters,
    generic_filter,
    generic_filters_validate,
)

__all__ = [
    "Filter",
    "CallableErrorHandler",
    "generic_filter",
    "generic_apply_filters",
    "generic_filters_validate",
]



from typing import Callable

import structlog
from sqlalchemy.inspection import inspect

from orchestrator.db import ProductTable, WorkflowTable
from orchestrator.db.database import SearchQuery
from orchestrator.db.filters import generic_filter
from orchestrator.db.filters.generic_filters import generic_is_like_filter, generic_range_filters
from orchestrator.utils.helpers import to_camel

logger = structlog.get_logger(__name__)

created_at_range_filters = generic_range_filters(WorkflowTable.created_at)


def products_filter(query: SearchQuery, value: str) -> SearchQuery:
    """Filter ProductBlocks by '-'-separated list of Product 'name' (column) values."""
    products = value.split("-")
    return query.filter(WorkflowTable.products.any(ProductTable.name.in_(products)))


BASE_CAMEL = {to_camel(key): generic_is_like_filter(value) for key, value in inspect(WorkflowTable).columns.items()}

WORKFLOW_FILTER_FUNCTIONS_BY_COLUMN: dict[str, Callable[[SearchQuery, str], SearchQuery]] = (
    BASE_CAMEL | {"products": products_filter} | created_at_range_filters
)

filter_workflows = generic_filter(WORKFLOW_FILTER_FUNCTIONS_BY_COLUMN)



# Copyright 2019-2020 SURF.
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import re
import subprocess  # noqa: S404
from typing import Generator, Optional

import structlog

from orchestrator.cli.generator.generator.helpers import get_variable

logger = structlog.getLogger(__name__)


def create_migration_file(config: dict) -> Optional[str]:
    create_migration_command = f'PYTHONPATH=. DATABASE_URI=postgresql://nwa:nwa@localhost/nwa-workflows python main.py db revision --message "Add {config["name"]}" --head=data@head'
    result = subprocess.check_output(create_migration_command, shell=True, text=True)  # noqa: S602

    for line in result.splitlines():
        if m := re.search("migrations/versions/general/([^ ]+)", line):
            return m[1]
    else:
        return None


def extract_revision_info(content: list[str]) -> dict:
    def process() -> Generator:
        for line in content:
            if m := re.search("Revision ID: (.+)", line):
                yield "revision", m[1]
            if m := re.search("Revises: (.+)", line):
                yield "down_revision", m[1]
            if m := re.search("Create Date: (.+)", line):
                yield "create_date", m[1]

    return dict(process())


def generate_product_migration(context: dict) -> None:
    config = context["config"]
    environment = context["environment"]
    writer = context["writer"]

    if migration_file := create_migration_file(config):
        path = f"migrations/versions/general/{migration_file}"
        try:
            with open(path) as stream:
                original_content = stream.readlines()
        except FileNotFoundError:
            logger.error("Migration file not found", path=path)
        else:
            revision_info = extract_revision_info(original_content)
            variable = get_variable(config)

            template = environment.get_template("new_product_migration.j2")
            content = template.render(product=config, variable=variable, **revision_info)

            writer(path, content)

    else:
        logger.error("Could not create migration file")



from orchestrator.db.filters.generic_filters.bool_filter import generic_bool_filter
from orchestrator.db.filters.generic_filters.is_like_filter import generic_is_like_filter
from orchestrator.db.filters.generic_filters.range_filter import (
    RANGE_TYPES,
    convert_to_date,
    convert_to_int,
    generic_range_filter,
    generic_range_filters,
    get_filter_value_convert_function,
)
from orchestrator.db.filters.generic_filters.values_in_column_filter import generic_values_in_column_filter

__all__ = [
    "RANGE_TYPES",
    "convert_to_date",
    "convert_to_int",
    "get_filter_value_convert_function",
    "generic_range_filter",
    "generic_range_filters",
    "generic_is_like_filter",
    "generic_values_in_column_filter",
    "generic_bool_filter",
]



from typing import Iterable, List, Optional, Set, Tuple, TypeVar, Union

import structlog

from orchestrator.cli.helpers.print_helpers import print_fmt

logger = structlog.get_logger(__name__)

T = TypeVar("T")


def get_user_input(text: str, default: str = "", optional: bool = False) -> str:
    while True:
        answer = input(text)
        if answer or default:
            return answer.strip() if answer else default
        if optional:
            return default


def _enumerate_menu_keys(items: Union[List, Set]) -> List[str]:
    return [str(i + 1) for i in range(len(items))]


def _prompt_user_menu(
    options: Iterable[Tuple[str, T]], keys: Optional[List[str]] = None, repeat: bool = True
) -> Optional[T]:
    options_list = list(options)
    keys = keys or _enumerate_menu_keys(options_list)
    done = False
    while not done:
        for k, txt_v in zip(keys, options_list):
            print_fmt(f"{k}) {txt_v[0]}")
        choice = get_user_input("? ")
        if choice not in keys:
            print_fmt("Invalid choice")
            done = not repeat
        else:
            return options_list[keys.index(choice)][1]
    return None



# Copyright 2019-2020 SURF.
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from http import HTTPStatus
from typing import List
from uuid import UUID

from fastapi.param_functions import Body
from fastapi.routing import APIRouter
from sqlalchemy.orm import joinedload

from orchestrator.api.error_handling import raise_status
from orchestrator.api.models import delete, save, update
from orchestrator.db import ProductBlockTable, ProductTable
from orchestrator.schemas import ProductBlockBaseSchema
from orchestrator.schemas import ProductBlockEnrichedSchema as ProductBlockSchema

router = APIRouter()


@router.get("/", response_model=List[ProductBlockSchema])
def fetch() -> List[ProductBlockTable]:
    return ProductBlockTable.query.options(joinedload(ProductBlockTable.resource_types)).all()


@router.get("/{product_block_id}", response_model=ProductBlockSchema)
def product_block_by_id(product_block_id: UUID) -> ProductBlockTable:
    product_block = (
        ProductBlockTable.query.options(joinedload(ProductBlockTable.resource_types))
        .filter_by(product_block_id=product_block_id)
        .first()
    )
    if not product_block:
        raise_status(HTTPStatus.NOT_FOUND, f"Product block id {product_block_id} not found")
    return product_block


@router.post("/", response_model=None, status_code=HTTPStatus.NO_CONTENT)
def save_product_block(data: ProductBlockBaseSchema = Body(...)) -> None:
    return save(ProductBlockTable, data)


@router.put("/", response_model=None, status_code=HTTPStatus.NO_CONTENT)
def update_product_block(data: ProductBlockBaseSchema = Body(...)) -> None:
    return update(ProductBlockTable, data)


@router.delete("/{product_block_id}", response_model=None, status_code=HTTPStatus.NO_CONTENT)
def delete_product_block(product_block_id: UUID) -> None:
    products = ProductTable.query.filter(
        ProductTable.product_blocks.any(ProductBlockTable.product_block_id == product_block_id)
    ).all()
    if len(products) > 0:
        error_products = ", ".join(map(lambda product: product.name, products))
        raise_status(HTTPStatus.BAD_REQUEST, f"ProductBlock {product_block_id} is used in Products: {error_products}")

    return delete(ProductBlockTable, product_block_id)



from importlib.machinery import SourceFileLoader

import toml
from setuptools import setup

version = SourceFileLoader("__version__", "orchestrator/__init__.py").load_module()

setup_variables = toml.load("pyproject.toml")["tool"]["flit"]["metadata"]

setup(
    name=setup_variables["dist-name"],
    version=str(version.__version__),
    classifiers=setup_variables["classifiers"],
    author=setup_variables["author"],
    author_email=setup_variables["author-email"],
    packages=[setup_variables["module"]],
    install_requires=setup_variables["requires"],
    description="The Orchestrator core",
    long_description=setup_variables["description-file"],
)



#!/usr/bin/env python3

# Copyright 2019-2020 SURF.
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from typing import Any, Callable, Dict, List, Optional, Type

import sentry_sdk
import structlog
import typer
from deprecated import deprecated
from fastapi.applications import FastAPI
from fastapi_etag.dependency import add_exception_handler
from pydantic.json import ENCODERS_BY_TYPE
from sentry_sdk.integrations import Integration
from sentry_sdk.integrations.asyncio import AsyncioIntegration
from sentry_sdk.integrations.fastapi import FastApiIntegration
from sentry_sdk.integrations.redis import RedisIntegration
from sentry_sdk.integrations.sqlalchemy import SqlalchemyIntegration
from sentry_sdk.integrations.threading import ThreadingIntegration
from starlette.middleware.cors import CORSMiddleware
from starlette.middleware.sessions import SessionMiddleware
from starlette.responses import JSONResponse, Response

from nwastdlib.logging import ClearStructlogContextASGIMiddleware, initialise_logging
from orchestrator import __version__
from orchestrator.api.api_v1.api import api_router
from orchestrator.api.error_handling import ProblemDetailException
from orchestrator.cli.main import app as cli_app
from orchestrator.db import db, init_database
from orchestrator.db.database import DBSessionMiddleware
from orchestrator.distlock import init_distlock_manager
from orchestrator.domain import SUBSCRIPTION_MODEL_REGISTRY, SubscriptionModel
from orchestrator.exception_handlers import problem_detail_handler
from orchestrator.graphql import Mutation, Query, create_graphql_router, register_domain_models
from orchestrator.services.processes import ProcessDataBroadcastThread
from orchestrator.settings import AppSettings, ExecutorType, app_settings
from orchestrator.utils.vlans import VlanRanges
from orchestrator.version import GIT_COMMIT_HASH
from orchestrator.websocket import init_websocket_manager
from pydantic_forms.exception_handlers.fastapi import form_error_handler
from pydantic_forms.exceptions import FormException

logger = structlog.get_logger(__name__)

sentry_integrations: List[Integration] = [
    SqlalchemyIntegration(),
    RedisIntegration(),
    FastApiIntegration(),
    AsyncioIntegration(),
    ThreadingIntegration(propagate_hub=True),
]


class OrchestratorCore(FastAPI):
    def __init__(
        self,
        title: str = "The Orchestrator",
        description: str = "The orchestrator is a project that enables users to run workflows.",
        openapi_url: str = "/api/openapi.json",
        docs_url: str = "/api/docs",
        redoc_url: str = "/api/redoc",
        version: str = __version__,
        default_response_class: Type[Response] = JSONResponse,
        base_settings: AppSettings = app_settings,
        **kwargs: Any,
    ) -> None:
        self.base_settings = base_settings
        websocket_manager = init_websocket_manager(base_settings)
        distlock_manager = init_distlock_manager(base_settings)
        self.broadcast_thread = ProcessDataBroadcastThread(websocket_manager)

        startup_functions: list[Callable] = [distlock_manager.connect_redis]
        shutdown_functions: list[Callable] = [distlock_manager.disconnect_redis]
        if websocket_manager.enabled:
            startup_functions.extend([websocket_manager.connect_redis, self.broadcast_thread.start])
            shutdown_functions.extend(
                [websocket_manager.disconnect_all, self.broadcast_thread.stop, websocket_manager.disconnect_redis]
            )

        # Make sure encoding is done correctly on the response.
        ENCODERS_BY_TYPE[VlanRanges] = str

        super().__init__(
            title=title,
            description=description,
            openapi_url=openapi_url,
            docs_url=docs_url,
            redoc_url=redoc_url,
            version=version,
            default_response_class=default_response_class,
            on_startup=startup_functions,
            on_shutdown=shutdown_functions,
            **kwargs,
        )

        initialise_logging()

        self.include_router(api_router, prefix="/api")

        init_database(base_settings)

        self.add_middleware(ClearStructlogContextASGIMiddleware)
        self.add_middleware(SessionMiddleware, secret_key=base_settings.SESSION_SECRET)
        self.add_middleware(DBSessionMiddleware, database=db)
        origins = base_settings.CORS_ORIGINS.split(",")
        self.add_middleware(
            CORSMiddleware,
            allow_origins=origins,
            allow_methods=base_settings.CORS_ALLOW_METHODS,
            allow_headers=base_settings.CORS_ALLOW_HEADERS,
            expose_headers=base_settings.CORS_EXPOSE_HEADERS,
        )

        self.add_exception_handler(FormException, form_error_handler)
        self.add_exception_handler(ProblemDetailException, problem_detail_handler)
        add_exception_handler(self)

        @self.router.get("/", response_model=str, response_class=JSONResponse, include_in_schema=False)
        def _index() -> str:
            return "Orchestrator orchestrator"

    @deprecated("Not using Opentelemetry from version 1.2.0, will be removed in 1.4")
    def instrument_app(self) -> None:
        pass

    def add_sentry(
        self,
        sentry_dsn: str,
        trace_sample_rate: float,
        server_name: str,
        environment: str,
        release: Optional[str] = GIT_COMMIT_HASH,
    ) -> None:
        logger.info("Adding Sentry middleware to app", app=self.title)
        if self.base_settings.EXECUTOR == ExecutorType.WORKER:
            from sentry_sdk.integrations.celery import CeleryIntegration

            sentry_integrations.append(CeleryIntegration())

        sentry_sdk.init(
            dsn=sentry_dsn,
            traces_sample_rate=trace_sample_rate,
            server_name=server_name,
            environment=environment,
            release=f"orchestrator@{release}",
            integrations=sentry_integrations,
            propagate_traces=True,
            profiles_sample_rate=trace_sample_rate,
        )

    @staticmethod
    def register_subscription_models(product_to_subscription_model_mapping: Dict[str, Type[SubscriptionModel]]) -> None:
        """Register your subscription models.

        This method is needed to register your subscription models inside the orchestrator core.

        Args:
            product_to_subscription_model_mapping: The dictionary should contain a mapping of products to SubscriptionModels.
                The selection will be done depending on the name of the product.

        Returns:
            None:

        Examples:
            >>> product_to_subscription_model_mapping = { # doctest:+SKIP
            ...     "Generic Product One": GenericProductModel,
            ...     "Generic Product Two": GenericProductModel,
            ... }

        """
        SUBSCRIPTION_MODEL_REGISTRY.update(product_to_subscription_model_mapping)

    def register_graphql(self: "OrchestratorCore", query: Any = Query, mutation: Any = Mutation) -> None:
        register_domain_models()
        new_router = create_graphql_router(query, mutation)
        self.include_router(new_router, prefix="/api/graphql")


main_typer_app = typer.Typer()
main_typer_app.add_typer(cli_app, name="orchestrator", help="The orchestrator CLI commands")

if __name__ == "__main__":
    main_typer_app()



from uuid import uuid4

import pytest
from sqlalchemy.exc import NoResultFound

from orchestrator.db import ProductTable
from orchestrator.services.products import get_product_by_id, get_product_by_name, get_tags, get_types


def test_get_product_by_id(generic_product_1):
    product = ProductTable.query.filter(ProductTable.name == "Product 1").one()

    result = get_product_by_id(product.product_id)
    assert result.product_id == product.product_id


def test_get_product_by_id_err(generic_product_1):
    assert get_product_by_id(uuid4()) is None


def test_get_product_by_name(generic_product_1):
    product = ProductTable.query.filter(ProductTable.name == "Product 1").one()

    result = get_product_by_name(product.name)
    assert result.product_id == product.product_id


def test_get_product_by_name_err(generic_product_1):
    with pytest.raises(NoResultFound):
        assert get_product_by_name("") is None


def test_get_types(generic_product_1):
    assert get_types() == ["Generic"]


def test_get_tags(generic_product_1):
    assert get_tags() == ["GEN1"]



from sqlalchemy.inspection import inspect

from orchestrator.db import SubscriptionTable
from orchestrator.db.sorting.product import PRODUCT_SORT_FUNCTIONS_BY_COLUMN
from orchestrator.db.sorting.sorting import generic_column_sort, generic_sort
from orchestrator.utils.helpers import to_camel

subscription_table_sort = {
    to_camel(key): generic_column_sort(value) for [key, value] in inspect(SubscriptionTable).columns.items()
}

SUBSCRIPTION_SORT_FUNCTIONS_BY_COLUMN = PRODUCT_SORT_FUNCTIONS_BY_COLUMN | subscription_table_sort

sort_subscriptions = generic_sort(SUBSCRIPTION_SORT_FUNCTIONS_BY_COLUMN)



# Copyright 2019-2020 SURF.
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from http import HTTPStatus

import structlog

from orchestrator.api.error_handling import raise_status
from orchestrator.db.database import SearchQuery

logger = structlog.get_logger(__name__)


def apply_range_to_query(query: SearchQuery, offset: int, limit: int) -> SearchQuery:
    """Apply range to the SearchQuery.

    Args:
        query: The sql query to add offset and limit to.
        offset: the limit item in the list to get.
        limit: the amount of items to get with offset as start point.

    returns the query with offset and limit applied.
    """

    if offset is not None and limit:
        if offset >= offset + limit:
            msg = "range start must be lower than end"
            logger.exception(msg)
            raise_status(HTTPStatus.BAD_REQUEST, msg)
        query = query.offset(offset).limit(limit + 1)
    return query



# Copyright 2019-2020 SURF.
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from graphql import GraphQLError

from orchestrator.db.filters import CallableErrorHandler
from orchestrator.graphql.extensions.error_collector_extension import register_error
from orchestrator.graphql.types import OrchestratorInfo


def create_resolver_error_handler(info: OrchestratorInfo) -> CallableErrorHandler:
    def _handle_error(message: str, **kwargs) -> None:  # type: ignore
        register_error(GraphQLError(message=message, path=info.path, extensions=(kwargs or {})))

    return _handle_error



import pytest
from pydantic import ValidationError

from orchestrator.forms.network_type_validators import BFD, MTU, VlanRangesValidator
from orchestrator.utils.vlans import VlanRanges
from pydantic_forms.core import FormPage


def test_bfd_ok():
    class Form(FormPage):
        bfd: BFD

    model = {"enabled": False, "minimum_interval": None, "multiplier": None}
    validated_data = Form(bfd=model).dict()

    expected = {"bfd": BFD(enabled=False, minimum_interval=None, multiplier=None)}
    assert expected == validated_data

    model = {"enabled": True, "minimum_interval": 600, "multiplier": 3}
    validated_data = Form(bfd=model).dict()

    expected = {"bfd": BFD(enabled=True, minimum_interval=600, multiplier=3)}
    assert expected == validated_data


def test_bfd_schema():
    class Form(FormPage):
        bfd: BFD

    assert Form.schema() == {
        "additionalProperties": False,
        "definitions": {
            "BFD": {
                "format": "optGroup",
                "properties": {
                    "enabled": {"title": "Enabled", "type": "boolean"},
                    "minimum_interval": {
                        "default": 900,
                        "maximum": 255000,
                        "minimum": 1,
                        "title": "Minimum " "Interval",
                        "type": "integer",
                    },
                    "multiplier": {
                        "default": 3,
                        "maximum": 255,
                        "minimum": 1,
                        "title": "Multiplier",
                        "type": "integer",
                    },
                },
                "required": ["enabled"],
                "title": "BFD",
                "type": "object",
            }
        },
        "properties": {"bfd": {"$ref": "#/definitions/BFD"}},
        "required": ["bfd"],
        "title": "unknown",
        "type": "object",
    }


def test_bfd_missing_values():
    class Form(FormPage):
        bfd: BFD

    model = {"enabled": False}
    assert Form(bfd=model).dict() == {"bfd": {"enabled": False}}

    model = {"enabled": True, "multiplier": 4}
    assert Form(bfd=model).dict() == {"bfd": {"enabled": True, "multiplier": 4, "minimum_interval": 900}}

    model = {"enabled": True, "minimum_interval": 600}
    assert Form(bfd=model).dict() == {"bfd": {"enabled": True, "multiplier": 3, "minimum_interval": 600}}

    model = {"enabled": False, "minimum_interval": 600, "multiplier": 3}
    assert Form(bfd=model).dict() == {"bfd": {"enabled": False}}


def test_bfd_wrong_values():
    class Form(FormPage):
        bfd: BFD

    with pytest.raises(ValidationError) as error_info:
        Form(bfd={"enabled": True, "minimum_interval": 0, "multiplier": 3})

    expected = [
        {
            "loc": ("bfd", "minimum_interval"),
            "msg": "ensure this value is greater than or equal to 1",
            "type": "value_error.number.not_ge",
            "ctx": {"limit_value": 1},
        },
    ]
    assert expected == error_info.value.errors()

    with pytest.raises(ValidationError) as error_info:
        Form(bfd={"enabled": True, "minimum_interval": 255001, "multiplier": 3})

    expected = [
        {
            "loc": ("bfd", "minimum_interval"),
            "msg": "ensure this value is less than or equal to 255000",
            "type": "value_error.number.not_le",
            "ctx": {"limit_value": 255000},
        },
    ]
    assert expected == error_info.value.errors()

    with pytest.raises(ValidationError) as error_info:
        Form(bfd={"enabled": True, "minimum_interval": 600, "multiplier": 0})

    expected = [
        {
            "loc": ("bfd", "multiplier"),
            "msg": "ensure this value is greater than or equal to 1",
            "type": "value_error.number.not_ge",
            "ctx": {"limit_value": 1},
        },
    ]
    assert expected == error_info.value.errors()

    with pytest.raises(ValidationError) as error_info:
        Form(bfd={"enabled": True, "minimum_interval": 600, "multiplier": 256})

    expected = [
        {
            "loc": ("bfd", "multiplier"),
            "msg": "ensure this value is less than or equal to 255",
            "type": "value_error.number.not_le",
            "ctx": {"limit_value": 255},
        },
    ]
    assert expected == error_info.value.errors()


def test_mtu():
    class Form(FormPage):
        mtu: MTU

    assert Form(mtu=1500).mtu == 1500
    assert Form(mtu=1501).mtu == 1501
    assert Form(mtu=9000).mtu == 9000


def test_mtu_schema():
    class Form(FormPage):
        mtu: MTU

    assert Form.schema() == {
        "additionalProperties": False,
        "properties": {
            "mtu": {"maximum": 9000, "minimum": 1500, "multipleOf": 7500, "title": "Mtu", "type": "integer"}
        },
        "required": ["mtu"],
        "title": "unknown",
        "type": "object",
    }


def test_mtu_nok():
    class Form(FormPage):
        mtu: MTU

    with pytest.raises(ValidationError) as error_info:
        assert Form(mtu=1499)

    expected = [
        {
            "ctx": {"limit_value": 1500},
            "loc": ("mtu",),
            "msg": "ensure this value is greater than or equal to 1500",
            "type": "value_error.number.not_ge",
        }
    ]
    assert expected == error_info.value.errors()

    with pytest.raises(ValidationError) as error_info:
        assert Form(mtu=9001)

    expected = [
        {
            "ctx": {"limit_value": 9000},
            "loc": ("mtu",),
            "msg": "ensure this value is less than or equal to 9000",
            "type": "value_error.number.not_le",
        },
    ]
    assert expected == error_info.value.errors()


def test_vlanranges():
    class Form(FormPage):
        vlanranges: VlanRangesValidator

    assert Form(vlanranges=4096).vlanranges == VlanRanges(4096)
    assert Form(vlanranges=0).vlanranges == VlanRanges(0)
    assert Form(vlanranges="0-4096").vlanranges == VlanRanges("0-4096")
    assert Form(vlanranges="1,2,3").vlanranges == VlanRanges("1,2,3")
    assert Form(vlanranges="1,1,1").vlanranges == VlanRanges("1")


def test_vlanranges_schema():
    class Form(FormPage):
        vlanranges: VlanRangesValidator

    assert Form.schema() == {
        "additionalProperties": False,
        "properties": {
            "vlanranges": {
                "examples": ["345", "20-23,45,50-100"],
                "format": "vlan",
                "pattern": "^([1-4][0-9]{0,3}(-[1-4][0-9]{0,3})?,?)+$",
                "title": "Vlanranges",
                "type": "string",
            }
        },
        "required": ["vlanranges"],
        "title": "unknown",
        "type": "object",
    }


def test_vlanranges_nok():
    class Form(FormPage):
        vlanranges: VlanRangesValidator

    with pytest.raises(ValidationError) as error_info:
        assert Form(vlanranges="4097")

    expected = [
        {"loc": ("vlanranges",), "msg": "4097 is out of range (0-4096).", "type": "value_error"},
    ]
    assert expected == error_info.value.errors()

    with pytest.raises(ValidationError) as error_info:
        assert Form(vlanranges=-1)

    expected = [
        {"loc": ("vlanranges",), "msg": "-1 is out of range (0-4096).", "type": "value_error"},
    ]
    assert expected == error_info.value.errors()

    with pytest.raises(ValidationError) as error_info:
        assert Form(vlanranges="a,b,c")

    expected = [
        {"loc": ("vlanranges",), "msg": "a,b,c could not be converted to a VlanRanges object.", "type": "value_error"},
    ]
    assert expected == error_info.value.errors()



# Copyright 2022-2023 SURF.
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from orchestrator.graphql.autoregistration import (
    EnumDict,
    add_class_to_strawberry,
    graphql_subscription_name,
    register_domain_models,
)
from orchestrator.graphql.schema import (
    GRAPHQL_MODELS,
    Mutation,
    OrchestratorGraphqlRouter,
    OrchestratorSchema,
    Query,
    create_graphql_router,
    custom_context_dependency,
    get_context,
    graphql_router,
)
from orchestrator.graphql.types import SCALAR_OVERRIDES

__all__ = [
    "GRAPHQL_MODELS",
    "SCALAR_OVERRIDES",
    "Query",
    "Mutation",
    "OrchestratorGraphqlRouter",
    "OrchestratorSchema",
    "custom_context_dependency",
    "get_context",
    "graphql_router",
    "create_graphql_router",
    "EnumDict",
    "add_class_to_strawberry",
    "graphql_subscription_name",
    "register_domain_models",
]



# Copyright 2019-2023 SURF.
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


from typing import Callable

from sqlalchemy import Column, String, cast

from orchestrator.db.database import SearchQuery


def generic_is_like_filter(field: Column) -> Callable[[SearchQuery, str], SearchQuery]:
    def like_filter(query: SearchQuery, value: str) -> SearchQuery:
        return query.filter(cast(field, String).ilike("%" + value + "%"))

    return like_filter
